# -*- coding: utf-8 -*-

# Local imports.
from etacad.globals import Direction, Orientation
from etacad.beam import Beam

# External imports.
import ezdxf
import pytest

from ezdxf.math import Vec3
from itertools import chain


@pytest.fixture
def beam_concrete():
    return Beam(width=0.4,
                height=0.8,
                length=12,
                x=-1,
                y=-1)


def test_attributes_beam_concrete(beam_concrete):
    # Geometric attributes.
    assert beam_concrete.width == .4
    assert beam_concrete.height == .8
    assert beam_concrete.length == 12
    assert beam_concrete.x == -1
    assert beam_concrete.y == -1

    # Concrete attributes.
    assert beam_concrete.concrete.volume == 0.4 * 0.8 * 12


def test_draw_longitudinal_beam_concrete(beam_concrete):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = beam_concrete.draw_longitudinal(document=doc, unifilar_bars=False)
    doc.saveas(filename="./tests/beam_concrete_longitudinal.dxf")

    # Concrete.
    assert len(entities["concrete"]["concrete_elements"]) == 4

    # Dimensions.
    assert len(entities["dimensions_elements"]) == 0


def test_draw_transverse_beam_concrete(beam_concrete):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = beam_concrete.draw_transverse(document=doc)
    doc.saveas(filename="./tests/beam_concrete_transverse.dxf")

    # Concrete.
    assert len(entities["concrete"]["concrete_elements"]) == 1


@pytest.fixture
def beam():
    return Beam(width=.2,
                height=.35,
                length=6,
                x=10,
                y=5,
                as_sup={.01: 3},
                as_right={.008: 2},
                as_inf={.016: 3},
                as_left={.008: 2},
                anchor_sup=.15,
                anchor_inf=.15,
                cover=.03,
                stirrups_db=.006,
                stirrups_sep=.15,
                stirrups_anchor=.1,
                columns=[[.2, .35], [.3, .35]],
                columns_pos=[0, 5.7],
                nomenclature="@",
                number_init=7)


def test_attributes_beam(beam):
    # Geometric attributes.
    assert beam.width == .2
    assert beam.height == .35
    assert beam.length == 6
    assert beam.x == 10
    assert beam.y == 5
    assert beam.direction == Direction.HORIZONTAL
    assert beam.orientation == Orientation.RIGHT

    # Steel attributes.
    assert beam.as_sup == {.01: 3}
    assert beam.as_right == {.008: 2}
    assert beam.as_inf == {.016: 3}
    assert beam.as_left == {.008: 2}
    assert beam.anchor_sup == [.15, .15, .15]
    assert beam.anchor_inf == [.15, .15, .15]
    assert beam.cover == .03

    # Stirrups attributes.
    assert beam.stirrups_db == [.006]
    assert beam.stirrups_sep == [.15]
    assert beam.stirrups_anchor == [.1]
    assert beam.stirrups_length == [5.24]
    assert beam.stirrups_x == [0.38]

    assert beam.stirrups[0].width == .16800000000000004
    assert beam.stirrups[0].x == 10.38
    assert beam.stirrups[0].y == 5.016

    # Columns attributes.
    assert beam.columns == [[0.2, 0.35], [0.3, 0.35]]
    assert beam.columns_pos == [0, 5.7]

    # Others.
    assert beam.nomenclature == "@"
    assert beam.number_init == 11


def test_draw_longitudinal_beam(beam):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = beam.draw_longitudinal(document=doc, x=2, y=3, unifilar_bars=False)
    doc.saveas(filename="./tests/beam_longitudinal.dxf")

    # Concrete.
    assert len(entities["concrete"]["concrete_elements"]) == 4

    # Dimensions.
    assert len(entities["dimensions_elements"]) == 1

    # Stirrups.
    assert len(entities["stirrups"][0]["steel_elements"]) == 35


def test_draw_transverse_beam(beam):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = beam.draw_transverse(document=doc, x=2, y=3)
    doc.saveas(filename="./tests/beam_transverse.dxf")

    # Concrete shape.
    assert [*entities["concrete"]["concrete_elements"][0].vertices()] == [(2,  3), (2, 3.35), (2.2, 3.35), (2.2, 3), (2, 3)]

    # Bars.
    assert entities["bars"][0]["steel_elements"][0].dxf.center == Vec3(2.0269999999999997, 3.32, 0)  # Top bar.
    assert entities["bars"][1]["steel_elements"][0].dxf.center == Vec3(2.1, 3.32, 0)  # Top bar.
    assert entities["bars"][2]["steel_elements"][0].dxf.center == Vec3(2.173, 3.32, 0)  # Top bar.
    assert entities["bars"][3]["steel_elements"][0].dxf.center == Vec3(2.174, 3.1266666666666665, 0)  # Right bar.
    assert entities["bars"][4]["steel_elements"][0].dxf.center == Vec3(2.174, 3.2233333333333333, 0)  # Right bar.
    assert entities["bars"][5]["steel_elements"][0].dxf.center == Vec3(2.03, 3.03, 0)  # Bottom bar.
    assert entities["bars"][6]["steel_elements"][0].dxf.center == Vec3(2.1, 3.03, 0)  # Bottom bar.
    assert entities["bars"][7]["steel_elements"][0].dxf.center == Vec3(2.17, 3.03, 0)  # Bottom bar.
    assert entities["bars"][8]["steel_elements"][0].dxf.center == Vec3(2.026, 3.1266666666666665, 0)  # Left bar.
    assert entities["bars"][9]["steel_elements"][0].dxf.center == Vec3(2.026, 3.2233333333333333, 0)  # Left bar.

    # Stirrups.
    assert entities["stirrups"][0]["steel_elements"][0].dxf.start == Vec3(2.0269999999999997, 3.3249999999999997, 0)
    assert entities["stirrups"][0]["steel_elements"][0].dxf.end == Vec3(2.173, 3.3249999999999997, 0)
    assert entities["stirrups"][0]["steel_elements"][1].dxf.start == Vec3(2.03, 3.022, 0)
    assert entities["stirrups"][0]["steel_elements"][1].dxf.end == Vec3(2.17, 3.022, 0)

    # General.
    assert len(entities["concrete"]["all_elements"]) == 3
    assert len(list(chain(*[bar_dict["all_elements"] for bar_dict in entities["bars"]]))) == 10
    assert len(list(chain(*[stirrup_dict["all_elements"] for stirrup_dict in entities["stirrups"]]))) == 22
    assert len(entities["all_elements"]) == 35


def test_draw_longitudinal_rebar_detailing_beam(beam):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = beam.draw_longitudinal_rebar_detailing(document=doc, x=-10, y=1, unifilar=False)
    doc.saveas(filename="./tests/beam_longitudinal_rebar_detailing.dxf")

    assert entities["text_elements"][0].dxf.align_point == Vec3(-10.56, 0.8175000000000001, 0)
    assert entities["barline_elements"][0].dxf.start == Vec3(-10.7, 0.43, 0.0)

    # General.
    assert len(entities["text_elements"]) == 4
    assert len(entities["bars"]) == 4
    assert len(entities["barline_elements"]) == 3
    assert len(entities["all_elements"]) == 53


def test_draw_transverse_rebar_detailing_beam(beam):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = beam.draw_transverse_rebar_detailing(document=doc, x=10, y=-5)
    doc.saveas(filename="./tests/beam_transverse_rebar_detailing.dxf")

    # Steel elements.
    assert entities["stirrups"][0]["steel_elements"][0].dxf.start == Vec3(10.011000000000001, -4.691, 0.0)

    # General.
    assert len(entities["stirrups"][0]["steel_elements"]) == 22
    assert len(entities["stirrups"][0]["dimensions_elements"]) == 4
    assert len(entities["all_elements"]) == 26


def test_draw_table_rebar_detailing_beam(beam):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = beam.draw_table_rebar_detailing(document=doc, x=-20, y=-5)
    doc.saveas(filename="./tests/beam_table_rebar_detailing.dxf")

    # General.
    assert len(entities["all_elements"]) == 66



# -*- coding: utf-8 -*-

# Imports.
# Local imports.
from etacad.globals import Aligment, Direction

# External imports.
import ezdxf
from ezdxf.gfxattribs import GfxAttribs
from ezdxf.document import Drawing
from ezdxf.transform import inplace
from ezdxf.math import Matrix44
from ezdxf.enums import TextEntityAlignment
from ezdxf.tools.standards import linetypes

from math import pi

# Classes references.
doc_class = Drawing
attrib_class = GfxAttribs

linetypes_list = linetypes(scale=0.1)
lt_center = linetypes_list[1]

# Matrix transformation definition.
matrix_x_mirror = Matrix44([1, 0, 0, 0],
                           [0, -1, 0, 0],
                           [0, 0, 1, 0],
                           [0, 0, 0, 1])

matrix_y_mirror = Matrix44([-1, 0, 0, 0],
                           [0, 1, 0, 0],
                           [0, 0, 1, 0],
                           [0, 0, 0, 1])


# Function that draws a circunference.
def circle(doc: Drawing,
           center_point: tuple,
           radius: float,
           attr=None):
    """
    Draws a circle in the provided drawing.

    :param doc: The drawing object where the circle will be drawn.
    :type doc: Drawing
    :param center_point: The center point of the circle as a tuple (x, y).
    :type center_point: tuple
    :param radius: The radius of the circle.
    :type radius: float
    :param attr: Optional DXF attributes for the circle.
    :type attr: dict, optional
    :return: A list containing the circle entity.
    :rtype: list
    """
    if not isinstance(doc, Drawing):
        return []

    msp = doc.modelspace()

    # Drawing of circle.
    circ = [msp.add_circle(center_point, radius=radius, dxfattribs=attr)]

    return circ


# Function that draws a curve with/out thickness.
def curve(doc: Drawing,
          center_point: tuple,
          radius: float,
          start_angle: float,
          end_angle: float,
          thickness: float = 0,
          attr=None) -> list:
    """
    Draws an arc with optional thickness in the provided drawing.

    :param doc: The drawing object where the arc will be drawn.
    :type doc: Drawing
    :param center_point: The center point of the arc as a tuple (x, y).
    :type center_point: tuple
    :param radius: The radius of the arc.
    :type radius: float
    :param start_angle: The starting angle of the arc in degrees.
    :type start_angle: float
    :param end_angle: The ending angle of the arc in degrees.
    :type end_angle: float
    :param thickness: Optional thickness of the arc. If specified, another arc with increased radius is drawn.
    :type thickness: float, optional
    :param attr: Optional DXF attributes for the arc.
    :type attr: dict, optional
    :return: A list containing the arc entities.
    :rtype: list
    """
    if not isinstance(doc, Drawing):
        return []

    msp = doc.modelspace()

    # Drawing of arc.
    arcs = [msp.add_arc(center_point, radius=radius, start_angle=start_angle, end_angle=end_angle, dxfattribs=attr)]

    if thickness:
        arcs += [msp.add_arc(center_point, radius=radius + thickness, start_angle=start_angle, end_angle=end_angle,
                             dxfattribs=attr)]

    return arcs


# Create a delimit axe and rerturns and list with elements.
def delimit_axe(document: Drawing,
                x: float,
                y: float,
                height: float = 1,
                radius: float = 0.1,
                text_height: float = 0.1,
                symbol: str = None,
                direction: Direction = Direction.VERTICAL,
                attr: GfxAttribs = None) -> list:
    """
    Creates a delimit axis with optional symbol in the specified drawing.

    :param document: The drawing object where the axis will be created.
    :type document: Drawing
    :param x: The x-coordinate of the starting point of the axis.
    :type x: float
    :param y: The y-coordinate of the starting point of the axis.
    :type y: float
    :param height: The height of the axis.
    :type height: float, optional
    :param radius: The radius of the symbol circle.
    :type radius: float, optional
    :param text_height: The height of the symbol text.
    :type text_height: float, optional
    :param symbol: The symbol to be placed at the end of the axis.
    :type symbol: str, optional
    :param direction: The direction of the axis (VERTICAL or HORIZONTAL).
    :type direction: Direction
    :param attr: Optional DXF attributes for the axis elements.
    :type attr: GfxAttribs, optional
    :return: A list containing the axis elements.
    :rtype: list
    """
    document.linetypes.add(name=lt_center[0], description=lt_center[1], pattern=lt_center[2])
    msp = document.modelspace()

    if direction == Direction.VERTICAL:
        group = [msp.add_line(start=(x, y), end=(x, y + height), dxfattribs=attr)]
        if symbol is not None:
            group += [msp.add_circle(center=(x, y + height + radius), radius=radius)]
            group += [msp.add_text(text=symbol, height=text_height).set_placement(
                p1=(x, y + height + radius), align=TextEntityAlignment.MIDDLE_CENTER)]

    elif direction == Direction.HORIZONTAL:
        group = [msp.add_line(start=(x + radius, y), end=(x + radius + height, y), dxfattribs=attr)]
        if symbol is not None:
            group += [msp.add_circle(center=(x, y), radius=radius)]
            group += [msp.add_text(text=symbol, height=text_height).set_placement(
                p1=(x, y), align=TextEntityAlignment.MIDDLE_CENTER)]

    else:
        raise ValueError

    return group


# Draws a linear dimension.
def dim_linear(document: Drawing,
               p_base: tuple,
               p1: tuple,
               p2: tuple,
               rotation: float = 0,
               dimstyle: str = "EZ_M_25_H25_CM",
               attr: GfxAttribs = None) -> list:
    """
    Draws a linear dimension between two points in the provided drawing.

    :param document: The drawing object where the dimension will be drawn.
    :type document: Drawing
    :param p_base: The base point of the dimension.
    :type p_base: tuple
    :param p1: The first point of the dimension.
    :type p1: tuple
    :param p2: The second point of the dimension.
    :type p2: tuple
    :param rotation: Optional rotation angle of the dimension in degrees.
    :type rotation: float, optional
    :param dimstyle: The dimension style to use.
    :type dimstyle: str, optional
    :param attr: Optional DXF attributes for the dimension.
    :type attr: GfxAttribs, optional
    :return: A list containing the dimension entity.
    :rtype: list
    """
    msp = document.modelspace()

    dims = [msp.add_linear_dim(base=p_base, p1=p1, p2=p2, dimstyle=dimstyle, angle=rotation, dxfattribs=attr)]
    dims[-1].render()

    return dims


# Function that draws a line.
def line(doc: Drawing, p1: tuple, p2: tuple, attr=None) -> list:
    """
    Draws a line between two points in the provided drawing.

    :param doc: The drawing object where the line will be drawn.
    :type doc: Drawing
    :param p1: The starting point of the line.
    :type p1: tuple
    :param p2: The ending point of the line.
    :type p2: tuple
    :param attr: Optional DXF attributes for the line.
    :type attr: dict, optional
    :return: A list containing the line entity.
    :rtype: list
    """
    line_segment = doc.modelspace().add_line(p1, p2, dxfattribs=attr)

    return [line_segment]


# Function that mirrors a given group of objects.
def mirror(objects: list, mirror_type: str, in_situ: bool = False) -> int:
    """
    Mirrors a group of objects along the specified axis.

    :param objects: A list of objects to be mirrored.
    :type objects: list
    :param mirror_type: The axis to mirror along, e.g., "x", "y".
    :type mirror_type: str
    :param in_situ: Whether to apply the mirror transformation in-place.
    :type in_situ: bool, optional
    :return: An integer status code indicating success (1).
    :rtype: int
    """
    # Axis "x" mirror.
    if "x" in mirror_type:
        inplace(objects, matrix_x_mirror)

    # Axis "y" mirror.
    if "y" in mirror_type:
        inplace(objects, matrix_y_mirror)

    return 1


def mtext(document: Drawing, textstr: str, height: float, x: float, y: float, width: float = 0, rotation: float = 0,
          aligment: int | Aligment = Aligment.MTEXT_TOP_LEFT.value, attrs: GfxAttribs = None):
    msp = document.modelspace()

    multi_text = msp.add_mtext(text=textstr, dxfattribs=attrs).set_location(insert=(x, y),
                                                                            rotation=rotation,
                                                                            attachment_point=aligment)

    multi_text.dxf.char_height = height
    multi_text.dxf.width = width

    return [multi_text]


def polyline(document: Drawing, vertices: list, closed: bool = True) -> list:
    """
    Create a 2D polyline in the given DXF document with an option to close the polyline.

    :param document: The DXF document where the polyline will be added.
    :type document: ezdxf.document.Drawing
    :param vertices: A list of tuples representing the (x, y) coordinates of the polyline vertices.
    :type vertices: list[tuple[float, float]]
    :param closed: If True, the polyline will be closed (the last vertex connects to the first).
    :type closed: bool
    :return: A list containing the polyline entity created in the DXF modelspace.
    :rtype: list[ezdxf.entities.LWPolyline]

    :Example:

    >>> doc = ezdxf.new(dxfversion='R2010')
    >>> vertices = [(0, 0), (10, 0), (10, 10), (0, 10)]
    >>> pline = polyline(doc, vertices, closed=True)

    This function adds a lightweight 2D polyline to the modelspace of the provided DXF document.
    If the `closed` parameter is set to True, the polyline will loop back to the first vertex.
    """
    if closed:
        vertices = vertices + vertices[:1]

    msp = document.modelspace()

    pl = msp.add_lwpolyline(vertices)

    return [pl]


def rads(degrees: float) -> float:
    """
    Converts degrees to radians.

    :param degrees: The angle in degrees.
    :type degrees: float
    :return: The angle in radians.
    :rtype: float
    """
    return degrees * pi / 180


# Function that draws a rectangle in dxf file given.
def rect(doc: Drawing, width: float, height: float, x: float, y: float, sides=None, fill: bool = False,
         polyline: bool = False, attr: GfxAttribs = None) -> list:
    """
    Draws a rectangle in the provided drawing.

    :param doc: The drawing object where the rectangle will be drawn.
    :type doc: Drawing
    :param width: The width of the rectangle.
    :type width: float
    :param height: The height of the rectangle.
    :type height: float
    :param x: The x-coordinate of the bottom-left corner of the rectangle.
    :type x: float
    :param y: The y-coordinate of the bottom-left corner of the rectangle.
    :type y: float
    :param sides: A list of four boolean values indicating which sides of the rectangle should be drawn.
    :type sides: list, optional
    :param fill: Whether to fill the rectangle with a hatch pattern.
    :type fill: bool, optional
    :param polyline: Whether to draw the rectangle as a polyline.
    :type polyline: bool, optional
    :param attr: Optional DXF attributes for the rectangle.
    :type attr: dict, optional
    :return: A list containing the rectangle elements.
    :rtype: list
    """
    if sides is None:
        sides = [1, 1, 1, 1]

    if not isinstance(doc, Drawing):
        return []

    msp = doc.modelspace()
    lines = []

    # Polyline rectangle.
    if polyline:
        lines.append(msp.add_lwpolyline([(x, y), (x + width, y), (x + width, y + height), (x, y + height), (x, y)],
                                        dxfattribs=attr))
    else:
        # Horizontal lines.
        if sides[0]:
            lines.append(msp.add_line((x, y + height), (x + width, y + height), dxfattribs=attr))  # Top side.
        if sides[2]:
            lines.append(msp.add_line((x, y), (x + width, y), dxfattribs=attr))  # Bottom side.

        # Vertical lines.
        if sides[1]:
            lines.append(msp.add_line((x + width, y), (x + width, y + height), dxfattribs=attr))  # Right side.
        if sides[3]:
            lines.append(msp.add_line((x, y), (x, y + height), dxfattribs=attr))  # Left side.

    if fill:
        hatch = msp.add_hatch(color=7, dxfattribs=attr)
        hatch.paths.add_polyline_path([(x, y), (x + width, y), (x + width, y + height), (x, y + height)])
        hatch.set_pattern_fill("ANSI31", scale=0.01)
        lines.append(hatch)

    return lines


# Function that draws a rectangle in dxf file given.
def rect_border_curve(doc: Drawing, width: float, height: float, radius: float, x: float, y: float,
                      thickness: float = 0, sides=None, curves=None, curves_radius=None, attr=None) -> list:
    """
    Draws a rectangle with optional border curves in the provided drawing.

    :param doc: The drawing object where the rectangle will be drawn.
    :type doc: Drawing
    :param width: The width of the rectangle.
    :type width: float
    :param height: The height of the rectangle.
    :type height: float
    :param radius: The radius of the corner curves.
    :type radius: float
    :param x: The x-coordinate of the bottom-left corner of the rectangle.
    :type x: float
    :param y: The y-coordinate of the bottom-left corner of the rectangle.
    :type y: float
    :param thickness: Optional thickness of the border.
    :type thickness: float, optional
    :param sides: A list of four boolean values indicating which sides of the rectangle should be drawn.
    :type sides: list, optional
    :param curves: A list of four boolean values indicating which corners should have curves.
    :type curves: list, optional
    :param curves_radius: A list of four values for the radii of the corner curves.
    :type curves_radius: list, optional
    :param attr: Optional DXF attributes for the rectangle.
    :type attr: dict, optional
    :return: A list containing the rectangle elements.
    :rtype: list
    """
    if sides is None:
        sides = [1, 1, 1, 1]

    if curves is None:
        curves = [1, 1, 1, 1]

    if curves_radius is None:
        curves_radius = [radius, radius, radius, radius]

    if not isinstance(doc, Drawing):
        return []

    msp = doc.modelspace()
    lines = []

    # Horizontal lines.
    if sides[0]:
        lines.append(msp.add_line(start=(x + curves_radius[0], y + height),
                                  end=(x - curves_radius[1] + width, y + height), dxfattribs=attr))  # Top side.
    if sides[2]:
        lines.append(msp.add_line(start=(x + curves_radius[3], y),
                                  end=(x - curves_radius[2] + width, y), dxfattribs=attr))  # Bottom side.

    # Vertical lines.
    if sides[1]:
        lines.append(msp.add_line(start=(x + width, y + curves_radius[2]),
                                  end=(x + width, y - curves_radius[1] + height), dxfattribs=attr))  # Right side.
    if sides[3]:
        lines.append(msp.add_line(start=(x, y + curves_radius[3]),
                                  end=(x, y - curves_radius[0] + height), dxfattribs=attr))  # Left side.

    if thickness:
        # Horizontal lines borders.
        if sides[0]:
            lines.append(msp.add_line(start=(x + curves_radius[0], y + thickness + height),
                                      end=(x - curves_radius[1] + width, y + thickness + height),
                                      dxfattribs=attr))  # Top side.
        if sides[2]:
            lines.append(msp.add_line(start=(x + curves_radius[3], y - thickness),
                                      end=(x - curves_radius[2] + width, y - thickness),
                                      dxfattribs=attr))  # Bottom side.

        # Vertical lines borders.
        if sides[1]:
            lines.append(msp.add_line(start=(x + thickness + width, y + curves_radius[2]),
                                      end=(x + thickness + width, y - curves_radius[1] + height),
                                      dxfattribs=attr))  # Right side.
        if sides[3]:
            lines.append(msp.add_line(start=(x - thickness, y + curves_radius[3]),
                                      end=(x - thickness, y - curves_radius[0] + height),
                                      dxfattribs=attr))  # Left side.

    # Curves.
    if (sides[0] or sides[3]) and curves[0]:
        lines += curve(doc=doc, center_point=(x + curves_radius[0], y - curves_radius[0] + height),
                       start_angle=90, end_angle=180, radius=curves_radius[0], thickness=thickness, attr=attr)
    if (sides[0] or sides[1]) and curves[1]:
        lines += curve(doc=doc, center_point=(x - curves_radius[1] + width, y - curves_radius[1] + height),
                       start_angle=0, end_angle=90,
                       radius=curves_radius[1], thickness=thickness, attr=attr)
    if (sides[1] or sides[2]) and curves[2]:
        lines += curve(doc=doc, center_point=(x - curves_radius[2] + width, y + curves_radius[2]), start_angle=270,
                       end_angle=360,
                       radius=curves_radius[2], thickness=thickness, attr=attr)
    if (sides[2] or sides[3]) and curves[3]:
        lines += curve(doc=doc, center_point=(x + curves_radius[3], y + curves_radius[3]), start_angle=180,
                       end_angle=270,
                       radius=curves_radius[3], thickness=thickness, attr=attr)

    return lines


# Function that rotates elements.
def rotate(objects: list, angle: float) -> int:
    """
    Rotates a group of objects by a specified angle.

    :param objects: A list of objects to be rotated.
    :type objects: list
    :param angle: The angle to rotate the objects, in degrees.
    :type angle: float
    :return: An integer status code indicating success (1).
    :rtype: int
    """

    ezdxf.transform.z_rotate(entities=objects, angle=angle)

    return 1


# Function that draws simple text.
def text(document: Drawing, text: str, height: float, point: tuple, rotation: float = 0, attr=None) -> list:
    """
    Draws text in the provided drawing.

    :param document: The drawing object where the text will be drawn.
    :type document: Drawing
    :param text: The text string to be drawn.
    :type text: str
    :param height: The height of the text.
    :type height: float
    :param point: The location of the text's insertion point.
    :type point: tuple
    :param rotation: Optional rotation angle of the text, in degrees.
    :type rotation: float, optional
    :param attr: Optional DXF attributes for the text.
    :type attr: dict, optional
    :return: A list containing the text entity.
    :rtype: list
    """
    msp = document.modelspace()

    group = msp.add_text(text=text, height=height, rotation=rotation, dxfattribs=attr).set_placement(point)

    return [group]


# Function that translates a group of elements.
def translate(objects: list, vector: tuple) -> int:
    """
    Translates a group of objects by a specified vector.

    :param objects: A list of objects to be translated.
    :type objects: list
    :param vector: The translation vector (dx, dy).
    :type vector: tuple
    :return: An integer status code indicating success (1).
    :rtype: int
    """
    ezdxf.transform.translate(entities=objects, offset=vector)

    return 1



# -*- coding: utf-8 -*-

# External imports.
from attrs import define, field


@define(slots=True)
class Element:
    """
    General proposes for Element objects.

    :param width: Width of the Element.
    :type width: float
    :param height: Height of the Element.
    :type height: float
    :param element_type: Type of the Element (bar, stirrup, beam, etc.).
    :type element_type: str
    :param x: X coordinate of the bottom left corner bounding box Element point.
    :type x: float
    :param y: Y coordinate of the bottom left corner bounding box Element point.

    :ivar width: Width of the Element.
    :ivar height: Height of the Element.
    :ivar element_type: Type of the Element (bar, stirrup, beam, etc.).
    :ivar x: X coordinate of the bottom left corner bounding box Element point.
    :ivar y: Y coordinate of the bottom left corner bounding box Element point.
    """

    # Geometric attributes.
    width: float
    height: float
    element_type: str
    x: float = field(default=0)
    y: float = field(default=0)



# -*- coding: utf-8 -*-

# Imports.
# Local imports.
from etacad.globals import DRotation
from .utils import get_linear_center, get_line_longitud, get_line_eq, segment_center

# External imports.
from attrs import define, field


@define
class Polygon:
    """
    Represents a polygon with vertices, perimeter, area, and centroid.

    :param vertices: A list of tuples representing the vertices of the polygon.
                     Each tuple contains two float values (x, y) for the coordinates.
    :type vertices: list[tuple[float, float]]

    :ivar vertices: A list of tuples representing the vertices of the polygon.
    :vartype vertices: list[tuple[float, float]]
    :ivar perimeter: The perimeter of the polygon.
    :vartype perimeter: float
    :ivar area: The area of the polygon.
    :vartype area: float
    :ivar centroid: The centroid of the polygon, represented as a tuple of two floats (x, y).
    :vartype centroid: tuple[float, float]
    """
    # Geometry.
    vertices: list[tuple[float, float]]
    perimeter: float = field(init=False, default=0.0)
    area: float = field(init=False, default=0.0)
    centroid: tuple[float, float] = field(init=False, default=(0.0, 0.0))

    def __attrs_post_init__(self):
        self.perimeter = get_line_longitud(self.vertices)
        self.area = self.get_area()
        self.centroid = get_linear_center(points=self.vertices)

    def get_area(self) -> float:
        """
        Calculates the area of the polygon using the Shoelace formula.

        :return: The area of the polygon.
        :rtype: float
        """
        area = 0.0
        n = len(self.vertices)
        for i in range(n):
            x1, y1 = self.vertices[i]
            x2, y2 = self.vertices[(i + 1) % n]
            area += x1 * y2 - x2 * y1

        return abs(area) / 2.0

    def get_top_point(self) -> list:
        """
        Get the topmost points from the vertices.

        This method finds all points in the list of vertices that have the maximum Y-coordinate (top points).
        It returns the list of points sorted by the X-coordinate.

        :return: A list of points with the highest Y-coordinate, sorted by X.
        :rtype: list
        """
        max_y = max(self.vertices, key=lambda vertice: vertice[1])[1]
        top_points = [point for point in self.vertices if point[1] == max_y]
        top_points.sort(key=lambda point: point[0])

        return top_points

    def get_right_point(self) -> list:
        """
        Get the rightmost points from the vertices.

        This method finds all points in the list of vertices that have the maximum X-coordinate (right points).
        It returns the list of points sorted by the X-coordinate.

        :return: A list of points with the highest X-coordinate, sorted by X.
        :rtype: list
        """
        max_x = max(self.vertices, key=lambda vertice: vertice[0])[0]
        right_points = [point for point in self.vertices if point[0] == max_x]
        right_points.sort(key=lambda point: point[1])

        return right_points

    def get_bottom_point(self) -> list:
        """
        Get the bottommost points from the vertices.

        This method finds all points in the list of vertices that have the minimum Y-coordinate (bottom points).
        It returns the list of points sorted by the X-coordinate.

        :return: A list of points with the lowest Y-coordinate, sorted by X.
        :rtype: list
        """
        min_y = min(self.vertices, key=lambda vertice: vertice[1])[1]
        bottom_points = [point for point in self.vertices if point[1] == min_y]
        bottom_points.sort(key=lambda point: point[0])

        return bottom_points

    def get_left_point(self) -> list:
        """
        Get the leftmost points from the vertices.

        This method finds all points in the list of vertices that have the minimum X-coordinate (left points).
        It returns the list of points sorted by the X-coordinate.

        :return: A list of points with the lowest X-coordinate, sorted by X.
        :rtype: list
        """
        min_x = min(self.vertices, key=lambda vertice: vertice[0])[0]
        left_points = [point for point in self.vertices if point[0] == min_x]
        left_points.sort(key=lambda point: point[1])

        return left_points

    def get_points_direction_of_rotation(self) -> DRotation:
        """
        Determine the direction of rotation based on the top point of the vertices.

        This method checks the relative positions of the neighboring points of the topmost point
        in the list of vertices. Based on their relative X-coordinate, it determines whether the
        direction of rotation is clockwise or counterclockwise.

        :return: Direction of rotation (clockwise or counterclockwise).
        :rtype: DRotation
        """
        top_point = self.get_top_point()[-1]
        top_index = self.vertices.index(top_point)

        if top_index - 1 > 0:
            if self.vertices[(top_index - 1)][0] <= top_point[0]:
                return DRotation.CLOCKWISE
            else:
                return DRotation.COUNTERCLOCKWISE
        else:
            if self.vertices[(top_index + 1)][0] >= top_point[0]:
                return DRotation.CLOCKWISE
            else:
                return DRotation.COUNTERCLOCKWISE

    def get_points_between_index_limits_by_dr(self,
                                              index1: int,
                                              index2: int,
                                              direction_rotation: DRotation = DRotation.CLOCKWISE):

        # Returning points.
        if self.get_points_direction_of_rotation() == DRotation.CLOCKWISE:
            if direction_rotation == DRotation.CLOCKWISE:
                if index1 <= index2:
                    return self.vertices[index1:index2 + 1]
                else:
                    return self.vertices[index1:] + self.vertices[:index2 + 1]
            else:
                if index1 <= index2:
                    return self.vertices[index1::-1] + self.vertices[-1:index2 - 1:-1]
                else:
                    return self.vertices[index2:index1:-1]
        else:
            if direction_rotation == DRotation.CLOCKWISE:
                if index1 <= index2:
                    return self.vertices[index1::-1] + self.vertices[:index2 - 1:-1]
                else:
                    return self.vertices[-index2:-index1 - 1:-1]
            else:
                if index1 <= index2:
                    return self.vertices[index1:index2 + 1]
                else:
                    return self.vertices[index1:] + self.vertices[:index2 + 1]

    def get_side_equations(self, index1: int = None, index2: int = None):
        if index1 is None:
            index1 = 0

        if index2 is None:
            index2 = len(self.vertices)

        sides = []
        for point1 in self.vertices[index1:index2]:
            if self.vertices.index(point1) == len(self.vertices) - 1:
                point2 = self.vertices[0]
            else:
                point2 = self.vertices[self.vertices.index(point1) + 1]

            x1, y1 = point1
            x2, y2 = point2
            a, b = get_line_eq(x1=x1, y1=y1, x2=x2, y2=y2)
            sides.append((a, b))

        return sides

    def get_side_centers(self, index1: int = None, index2: int = None):
        if index1 is None:
            index1 = 0

        if index2 is None:
            index2 = len(self.vertices)

        segment_centers = []
        for point1 in self.vertices[index1:index2]:
            if self.vertices.index(point1) == len(self.vertices) - 1:
                point2 = self.vertices[0]
            else:
                point2 = self.vertices[self.vertices.index(point1) + 1]

            x1, y1 = point1
            x2, y2 = point2

            segment_centers.append(segment_center([x1, y1, x2, y2]))

        return segment_centers



# -*- coding: utf-8 -*-

# Imports.
# Locals imports.

# Externals imports.
import math


def displace_perpendicular(x0, y0, a, d):
    """
    Calculates coordinates at perpendicular distance from slope, x and y coordinates given.

    :param x0: X coordinate.
    :type x0: float
    :param y0: Y coordinate.
    :type y0: float
    :param a: Slope of line.
    :type a: float
    :param d: Perpendicular distance.
    :type d: float
    :return: Coordiantes at perpendicular distance.
    :rtype: tuple
    """
    # Perpendicular unit vector.
    if a == 0:
        unit_vector = (0, 1)
    else:
        unit_vector = (1 / math.sqrt((1 / a) ** 2 + 1), (-1 / a) / math.sqrt((1 / a) ** 2 + 1))

    # New point displaced by D in perpendicular direction.
    x1 = x0 + d * unit_vector[0]
    y1 = y0 + d * unit_vector[1]

    return x1, y1


def get_euclidean_distance(point1: tuple | list, point2: tuple | list):
    """
    Calculate the Euclidean distance between two Cartesian coordinates.

    :param point1: Tuple or list containing the x and y coordinates of first point.
    :type point1: tuple | list
    :param point2: Tuple or list containing the x and y coordinates of first point.
    :type point2: tuple | list
    :return: Distance between the two points.
    :rtype: float
    """
    x1, y1 = point1
    x2, y2 = point2

    distance_between_points = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)

    return distance_between_points


def get_angle(point1: tuple | list, point2: tuple | list, degrees: bool = True, axis: str = "x") -> float:
    """
    Calculates de angle of to points, sorted by axis given (default X).

    :param point1: First point (x, y).
    :type point1: tuple | list
    :param point2: Second point (x, y).
    :type point2: tuple | list
    :param degrees: Transforms de angle from radians to degrees if it True.
    :type degrees: bool
    :param axis: String that specifies the axis by which the points are ordered.
    :type axis: str
    :return: angle in radians (default).
    :rtype: float
    """
    if axis == "x":
        point1, point2 = sort_points(point1, point2, axis="x")
    if axis == "y":
        point1, point2 = sort_points(point1, point2, axis="y")

    x1, y1, x2, y2 = point1[0], point1[1], point2[0], point2[1]

    if degrees:
        return math.atan2(y2 - y1, x2 - x1) * 180 / math.pi

    if not degrees:
        return math.atan2(y2 - y1, x2 - x1)


def get_angle_x_axis(point: tuple, p0: tuple = (0, 0), degrees: bool = True) -> float:
    """
    Calculates the angle of the given point from the X axis, the origin can be specified.

    :param point: Tuple of floats (x, y) representing the point to calculate angle from X axis.
    :type point: tuple
    :param p0: Point (x, y) that represents the origin.
    :type p0: tuple
    :param degrees: Transforms de angle from radians to degrees if it True.
    :type degrees: bool
    :return: Angle from X axis in degrees.
    :rtype: float
    """
    # Calculate the slope of the line.
    if point[0] != p0[0]:  # Avoid division by zero.
        slope = (point[1] - p0[1]) / (point[0] - p0[0])
    else:
        slope = math.inf

    # Calculate the angle with respect to the X-axis.
    if degrees:
        angle = math.atan(slope) * 180 / math.pi  # Convert radians to degrees.
    else:
        angle = math.atan(slope)

    # Adjust the angle if necessary.
    if slope < 0:
        if point[1] > p0[1]:
            angle += 180 if degrees else math.pi
        if point[1] < p0[1]:
            angle += 360 if degrees else math.pi * 2
    elif slope > 0:
        if point[1] < p0[1]:
            angle += 180 if degrees else math.pi
        elif point[1] == p0[1] and point[0] < p0[0]:
            angle += 180 if degrees else math.pi
    elif slope == 0:
        if point[0] < 0:
            angle += 180 if degrees else math.pi

    return angle


def get_line_longitud(line_points: list | tuple) -> float:
    """
    Calculate the total distance for a list of Cartesian coordinates.

    :param line_points: List of tuples where each tuple contains the x and y coordinates of a point.
    :type line_points: list | tuple.
    :return: Total distance between the points in the list.
    :rtype: float
    """

    length = 0.0
    for i in range(len(line_points) - 1):
        length += get_euclidean_distance(line_points[i], line_points[i + 1])

    return length


def get_linear_center(points: list) -> tuple:
    """
    Calculates the lienar center of a polygon given a list of points.

    :param points: List of tuples representing the polygon vertices [(x1, y1), (x2, y2), ..., (xn, yn)]
    :return: A tuple (xm, ym) representing the linear center coordinates
    """
    xm = sum([point[0] for point in points]) / len(points)
    ym = sum([point[1] for point in points]) / len(points)

    return xm, ym


def get_parallel_intercept(slope: float, intercep: float, distance: float) -> float:
    """
    Calculates the intercept of the parallel line of the given line (slope, intercep) at the distance given.

    :param slope: Slope of the line from which to obtain the parallel line intercep.
    :type slope: float
    :param intercep: Intercep of the line from which to obtain the parallel line intercep.
    :type intercep: float
    :param distance: Distance at which to obtain the parallel line intercep.
    :type distance: float
    :return: Parallalel line intercep at distance given.
    :rtype: float
    """
    return intercep + distance * math.sqrt(1 + slope ** 2)


def get_paralel_intersec_from_list_of_points(points: list[tuple], pn: tuple, d: float) -> list:
    """
    Calculates the intersections of the parallels lines formed by the points towards the given node point, in
    counterclockwise order.

    :param points: List of tuples of floats representing the points of the lines.
    :type points: list
    :param pn: Point representing the node (x, y).
    :type pn: tuple
    :param d: Perpendicular distance of the parallels lines to the points given.
    :type d: float
    :return: Intersections of the paralles lines in clockwise order.
    :rtype: list
    """
    xn, yn = pn
    slopes_intercepts = []
    parallel_slopes_intercepts = []

    def points_sort_function(point: tuple) -> float:
        return get_angle_x_axis(point=(point[0], point[1]), p0=(xn, yn))

    points_sorted = sorted(points, key=points_sort_function)  # Sorting the points in order of angle from X axis.

    for x, y in points_sorted:
        slope, intercep = get_line_eq(xn, yn, x, y, aprox_inf=True)
        slopes_intercepts.append((slope, intercep))

        if slope >= 0:
            if x >= xn and y >= yn:
                parallel_slopes_intercepts.append((slope, get_parallel_intercept(slope, intercep, -d)))
                parallel_slopes_intercepts.append((slope, get_parallel_intercept(slope, intercep, d)))
            else:
                parallel_slopes_intercepts.append((slope, get_parallel_intercept(slope, intercep, d)))
                parallel_slopes_intercepts.append((slope, get_parallel_intercept(slope, intercep, -d)))
        if slope < 0:
            if x <= xn:
                parallel_slopes_intercepts.append((slope, get_parallel_intercept(slope, intercep, d)))
                parallel_slopes_intercepts.append((slope, get_parallel_intercept(slope, intercep, -d)))
            else:
                parallel_slopes_intercepts.append((slope, get_parallel_intercept(slope, intercep, -d)))
                parallel_slopes_intercepts.append((slope, get_parallel_intercept(slope, intercep, d)))

    intersections = []
    num_lines = len(parallel_slopes_intercepts)

    if num_lines == 2:
        return []

    for i in range(1, num_lines + 1, 2):
        if parallel_slopes_intercepts[i][0] != parallel_slopes_intercepts[(i + 1) % num_lines][0]:
            intersection = get_lines_intersec(parallel_slopes_intercepts[i][0],
                                              parallel_slopes_intercepts[i][1],
                                              parallel_slopes_intercepts[(i + 1) % num_lines][0],
                                              parallel_slopes_intercepts[(i + 1) % num_lines][1])
        else:
            intersection = get_lines_intersec_at_distance(r1=parallel_slopes_intercepts[i],
                                                          r2=parallel_slopes_intercepts[(i + 1) % num_lines],
                                                          d=d,
                                                          x1=pn[0])
        intersections.append(intersection)

    return intersections


# Returns the a and b constants of the line equation from two poins.
def get_line_eq(x1, y1, x2, y2, aprox_inf: bool = True):
    """
    Calculates slope and intercep of the line formed by the two points given.

    :param x1: X coordinate of the first point.
    :type x1: float
    :param y1: Y coordinate of the first point.
    :type y1: float
    :param x2: X coordinate of the second point.
    :type x2: float
    :param y2: Y coordinate of the second point.
    :type y2: float
    :param aprox_inf: Boolean that specified if the vertical lines are aproximated by 9999999999999999 slope or not.
    :type aprox_inf: bool
    :return: tuple of float representing slope and intercep of line.
    :rtype: tuple
    """
    if x1 == x2:  # For vertical lines.
        a = float(9999999999999999) if aprox_inf else float("inf")  # Representation of inf. slope for a vertical line.
        b = - a * x1

    else:  # For not vertical lines.
        a = (y2 - y1) / (x2 - x1)
        b = y1 - a * x1

    return a, b


def get_normal_line_eq(a: float, b: float, aprox_inf: bool = True) -> tuple:
    """
    Calculates slope and intercep of the normal line of the slope and intercep given.

    :param a: Slope of line.
    :type a: float
    :param b: Intercep of line.
    :type b: float
    :param aprox_inf: Boolean that specified if the vertical lines are aproximated by 9999999999999999 slope or not.
    :type aprox_inf: bool
    :return: tuple of float representing slope and intercep of normal line.
    :rtype: tuple
    """
    # If the slope is 0, the normal will be vertical.
    if a == 0:
        a_normal = float(9999999999999999) if aprox_inf else float("inf")  # Representation of inf. slope.
    else:
        # Calculate the slope of the normal (negative reciprocal).
        a_normal = -1 / a

    # The y-intercept of the normal remains the same as that of the line.
    b_normal = b

    return a_normal, b_normal


def get_lines_intersec(a1: float, b1: float, a2: float, b2: float) -> tuple[float, float]:
    """
    Calculates the intersection of the two lines given.

    :param a1: Slope of the first line.
    :type a1: float
    :param b1: Intercep of the first line.
    :type b1: float
    :param a2: Slope of the second line.
    :type a2: float
    :param b2: Intercep of the second line.
    :type b2: float
    :return: Intersection of the two lines given.
    :rtype: tuple
    """
    # Returns the intersection of two lines.
    x = (b2 - b1) / (a1 - a2)
    return x, x * a1 + b1


def get_lines_intersec_at_distance(r1: tuple, r2: tuple, d: float, x1: float) -> tuple:
    """
    Calculates the interction of the parallel lines of first and second line given at distance at X value given.

    :param r1: Slope and intercep of first line.
    :type r1: tuple
    :param r2: Slope and intercep of second line.
    :type r2: tuple
    :param d: Normal distance at the first and second line given.
    :type d: float
    :param x1:
    :type x1: float
    :return: Intersection (xi, yi) of the parallel lines at normal distance d.
    :rtype: tuple
    """
    m1, b1 = r1
    m2, b2 = r2

    # Verify that R1 and R2 are parallel by comparing their slopes
    if m1 != m2:
        raise ValueError("Lines R1 and R2 are not parallel.")

    # Find y1 using the equation of R1.
    y1 = m1 * x1 + b1

    # Calculate the angle of the lines with respect to the x-axis.
    angle = math.atan(m1)

    # Calculate the horizontal and vertical components of the distance D.
    dx = d * math.cos(angle)
    dy = d * math.sin(angle)

    # Calculate the coordinates of p2.
    x2 = x1 + dx
    y2 = y1 + dy

    return x2, y2


def get_reverse_axes(points: list) -> tuple:
    """
    Reverse the axes of the point given [x1, y1, x2, y2...] to [y1, x1, y2, x2...].

    :param points: list of x, y coordinates.
    :type points: list
    :return: X and Y coordinates reversed.
    :rtype: list
    """
    # Inverts the X and Y coordinates from a list of points, [x1, y1, x2, y2] -> [y1, x1, y2, x2]
    reversed_points = []
    for i in range(int(len(points) / 2)):
        reversed_points = reversed_points + [points[i * 2 + 1], points[i * 2]]

    return tuple(reversed_points)


def is_in_range(x: float, x1: float, x2: float) -> bool:
    """
    Verifies if X coordinate is in range of X1 and X2 limits.
    :param x: X coordinate to verify range.
    :param x1: First X coordinate limit.
    :param x2: Second X coordinat limit.
    :return: State of range X coordinate given.
    :rtype: bool
    """
    # Returns 1 if the given X is in the range X1, X2.
    extension = sorted([x1, x2])
    return extension[0] <= x <= extension[1]


def parallel_points(p1: tuple, p2: tuple, d: float, sort: bool = True) -> tuple:
    """
    Finds the parallels points at the distance given in the both sides of the line formed by the two points given.

    :param p1: First point of the line.
    :type p1: tuple
    :param p2: Second point of the line.
    :type p2: tuple
    :param d: Distance at where the parralels line pass.
    :type d: float
    :param sort: Sort the points by X coordinate if is it True.
    :type sort: bool
    :return: Tuple of four points (tuples (x, y)) of the parralels lines.
    :rtype: tuple
    """
    # Sort points.
    if sort:
        p1, p2 = sort_points(p1, p2, axis="x")

    # Calculate the slope of the original line.
    m, _ = get_line_eq(p1[0], p1[1], p2[0], p2[1])

    # Calculate the horizontal and vertical displacement.
    delta_y = d / ((1 + m ** 2) ** 0.5)
    delta_x = -m * delta_y

    # Calculate the displaced points.
    displaced_point1 = (p1[0] + delta_x, p1[1] + delta_y)
    displaced_point2 = (p2[0] + delta_x, p2[1] + delta_y)
    displaced_point3 = (p1[0] - delta_x, p1[1] - delta_y)
    displaced_point4 = (p2[0] - delta_x, p2[1] - delta_y)

    return displaced_point1, displaced_point2, displaced_point3, displaced_point4


def polygon_area(points: list[tuple[float, float]]) -> float:
    """
    Calculate the area of a polygon using the shoelace formula.

    The function takes a list of points representing the vertices of a polygon
    in 2D space and returns the area of the polygon. The vertices should be
    provided in a sequential order, either clockwise or counterclockwise.

    :param points: A list of tuples, where each tuple contains two floats
                   representing the (x, y) coordinates of a vertex.
                   Example: [(x1, y1), (x2, y2), ..., (xn, yn)].
    :type points: list[tuple[float, float]]

    :return: The area of the polygon.
    :rtype: float

    :raises ValueError: If less than 3 points are provided.

    Example:

    >>> points = [(0, 0), (4, 0), (4, 3), (0, 3)]
    >>> area = polygon_area(points)
    >>> print(area)  # Output: 12.0

    The formula used is known as the shoelace theorem or Gauss's area formula.
    """
    n = len(points)
    area = 0

    for i in range(n):
        x1, y1 = points[i]
        x2, y2 = points[(i + 1) % n]  # Next vertex, wrapping around to the first at the end.
        area += x1 * y2 - y1 * x2
    return abs(area) / 2


def scale_coordinates(*args, scale_factor: float):
    """
    Scales the args given by the scale factor given.

    :param args: List of floats to scale.
    :type args: float
    :param scale_factor: Number to multiplicate the args given.
    :type scale_factor: float
    :return: List of args scaled.
    :rtype: list
    """
    scaled_coordinates = [(x * scale_factor, y * scale_factor) for x, y in args]

    return scaled_coordinates


# Function that returns center of a line segment.
def segment_center(points: list) -> tuple:
    """
    Calculates the mid-point coordinates of segment.

    :param points: List of two points of the segment of line.
    :type points: list
    :return: Coordinates of mid-point of segment.
    :rtype: tuple
    """
    return (points[0] + points[2]) / 2, (points[1] + points[3]) / 2


def sort_points(*args, axis: str = "x") -> list:
    """
    Sort points by axis given (x or y).

    :param args: List of points (x, y).
    :type args: list
    :param axis:
    :type axis: str
    :return:
    """
    # Extracting individual points from args.
    points = [arg for arg in args]

    # Function to sort points by x-coordinate.
    def sort_by_x(point):
        return point[0]

    # Function to sort points by y-coordinate.
    def sort_by_y(point):
        return point[1]

    # Sorting the points based on the specified axis
    if axis == "x":
        sorted_points = sorted(points, key=sort_by_x)
    elif axis == "y":
        sorted_points = sorted(points, key=sort_by_y)
    else:
        raise ValueError("Invalid axis. Axis must be 'x' or 'y'.")

    return sorted_points



# -*- coding: utf-8 -*-

# Imports.
# Local imports.
from etacad.drawing_utils import circle, curve, line, mirror, rads, rect, rotate, text, translate
from etacad.globals import Direction, ElementTypes, Orientation, STEEL_WEIGHT, BAR_SET_LONG, BAR_SET_TRANSVERSE

# External imports.
from attrs import define, field
from ezdxf.document import Drawing
from math import cos, sin, tan, pi


@define
class Bar:
    """
    Longitudinal bar element, computes geometrics and physics props and manages dxf drawing methods.

    :param reinforcement_length: Length of the stirrup reinforcement.
    :type reinforcement_length: float
    :param diameter: Diameter of stirrup bar.
    :type diameter: float
    :param x: X coordinate of the bottom left corner Stirrup bounding box point.
    :type x: float
    :param y: Y coordinate of the bottom left corner Stirrup bounding box point.
    :type y: float
    :param left_anchor: Left anchor length of stirrup.
    :type left_anchor: float
    :param right_anchor: Right anchor length of stirrup.
    :type right_anchor: float
    :param mandrel_radius: Mandrel radius of bar.
    :type mandrel_radius: float
    :param direction: Direction of the stirrup (Horizontal or Vertical).
    :type direction: Direction
    :param orientation: Orientation of the stirrup (top, right, down, left).
    :type orientation: Orientation
    :param bend_longitud: Bending longitud of the bar at center of bar.
    :type: float
    :param bend_angle: Bending declination angle.
    :type bend_angle: float
    :param bend_height: Bending height of the bar at center of bar.
    :type bend_height: float
    :param transverse_center: Transverse center of the drawing of cross-section.
    :type transverse_center: tuple

    :ivar reinforcement_length: Total length of the reinforcement bar.
    :ivar diameter: Diameter of the bar.
    :ivar radius: Radius of the bar (half of the diameter).
    :ivar length: Overall length of the bar, including bends and anchors.
    :ivar x: X coordinate of the bottom left corner of the bounding box.
    :ivar y: Y coordinate of the bottom left corner of the bounding box.
    :ivar direction: Direction of the bar (horizontal or vertical).
    :ivar orientation: Positioning of the bar (top, right, down, left).
    :ivar transverse_center: Coordinates of the transverse center of the cross-section.
    :ivar left_anchor: Length of the left anchor.
    :ivar right_anchor: Length of the right anchor.
    :ivar mandrel_radius: Radius of the mandrel used for bending.
    :ivar mandrel_radius_ext: External radius of the mandrel plus the bar's diameter.
    :ivar bend_longitud: Length of the bar bend at its center.
    :ivar bend_angle: Angle of the bar bend.
    :ivar bend_height: Height of the bar at the bend.
    :ivar bending_proyection: Projection length caused by the bar's bend.
    :ivar box_width: Width of the bounding box containing the bar.
    :ivar box_height: Height of the bounding box containing the bar.
    :ivar weight: Weight of the bar, calculated using a steel density of 7850 kg/m³.
    :ivar element_type: Type of the element (BAR).
    :ivar denomination: Optional denomination for the bar element.
    """
    # Geometric attributes.
    reinforcement_length: float
    diameter: float
    radius: float = field(init=False)
    length: float = field(init=False)
    x: float = field(default=0)
    y: float = field(default=0)
    direction: Direction = field(default=Direction.HORIZONTAL)
    orientation: Orientation = field(default=Orientation.BOTTOM)
    transverse_center: tuple = field(default=None)

    # Anchor attributes.
    left_anchor: float = field(default=0)
    right_anchor: float = field(default=0)
    mandrel_radius: float = field(default=0)
    mandrel_radius_ext: float = field(init=False)

    # Bending attributes.
    bend_longitud: float = field(default=0)
    bend_angle: float = field(default=0)
    bend_height: float = field(default=0)
    bending_proyection: float = field(init=False)

    # Boxing attributes.
    box_width: float = field(init=False)
    box_height: float = field(init=False)

    # Physics attributes.
    weight: float = field(init=False)

    # Others.
    element_type: ElementTypes = field(default=ElementTypes.BAR)
    denomination: str = field(default=None)
    position: str = field(default=None)
    quantity: float = field(default=None)

    def __attrs_post_init__(self):
        # Bending attributes.
        self.bending_proyection = self.bend_height / tan(rads(self.bend_angle)) if self.bend_angle else 0

        # Geometric attributes.
        self.radius = self.diameter / 2
        self.length = (self.reinforcement_length + (1 / cos(rads(self.bend_angle)) - 1) * self.bending_proyection * 2
                       + self.left_anchor + self.right_anchor)

        # Mandrel attributes.
        self.mandrel_radius_ext = self.diameter + self.mandrel_radius

        # Boxing attributes.
        self.box_width = self.reinforcement_length
        if self.left_anchor or self.right_anchor or self.bend_height:
            max_anchor = max([self.left_anchor, self.right_anchor])  # Maximum anchor.

            self.box_height = max([self.mandrel_radius_ext + max_anchor, self.diameter * 2 + self.bend_height])
        else:
            self.box_height = self.diameter

        # Physics attributes.
        self.weight = (self.diameter ** 2 * pi / 4) * self.length * STEEL_WEIGHT

    # Drawing longitudinal function.
    def draw_longitudinal(self,
                          document: Drawing,
                          x: float = None,
                          y: float = None,
                          unifilar: bool = False,
                          dimensions: bool = True,
                          denomination: bool = True,
                          settings: dict = BAR_SET_LONG) -> dict:
        """
        Draws the longitudinal view of the bar in a DXF document.

        :param document: The DXF document to draw on.
        :type document: Drawing
        :param x: X coordinate for the drawing, defaults to self.x.
        :type x: float, optional
        :param y: Y coordinate for the drawing, defaults to self.y.
        :type y: float, optional
        :param unifilar: Whether to draw a unifilar representation (simplified view), defaults to False.
        :type unifilar: bool, optional
        :param dimensions: Whether to include dimensions in the drawing, defaults to True.
        :type dimensions: bool, optional
        :param denomination: Whether to include the denomination label, defaults to True.
        :type denomination: bool, optional
        :param settings: Dictionary of settings for dimensioning. Defaults to `BAR_SET_LONG`.
        :type settings: dict, optional
        :return: Dict of drawing entities for the longitudinal view.
        :rtype: dict
        """
        if x is None:
            x = self.x
        if x is None:
            y = self.y

        if not unifilar:
            diameter = self.diameter
            mandrel_radius_ext = self.mandrel_radius_ext
            sides_left_anchor = [0, 1, 1, 1]
            sides_right_anchor = [0, 1, 1, 1]
        else:
            diameter = 0
            mandrel_radius_ext = 0
            sides_left_anchor = [0, 0, 0, 1]
            sides_right_anchor = [0, 1, 0, 0]

        elements = {}
        steel_elements = []
        dimension_elements = []
        denomination_elements = []

        # Drawing of simple entities and setting of variables for progressive draw.
        length_first_rect_bar = self.reinforcement_length
        x_first_rect_bar = x
        sides_first_rect_bar = [1, 1, 1, 1]

        length_third_rect_bar = self.reinforcement_length
        sides_third_rect_bar = [1, 1, 1, 0]

        # Left anchor.
        if self.left_anchor:
            x_lab = x
            y_lab = y + self.box_height - mandrel_radius_ext - self.left_anchor
            center_point_lac = (x + mandrel_radius_ext,
                                y + self.box_height - mandrel_radius_ext)

            length_first_rect_bar -= mandrel_radius_ext
            x_first_rect_bar += mandrel_radius_ext
            sides_first_rect_bar[3] = 0

            # From left to right.
            # First anchor rect bar (left).
            steel_elements += rect(doc=document,
                                   width=diameter,
                                   height=self.left_anchor, x=x_lab, y=y_lab,
                                   sides=sides_left_anchor)

            if not unifilar:
                # First bend curve.
                steel_elements += curve(doc=document,
                                        center_point=center_point_lac,
                                        radius=self.mandrel_radius,
                                        start_angle=90,
                                        end_angle=180,
                                        thickness=diameter)

        # Right anchor.
        if self.right_anchor:
            x_rab = x + self.reinforcement_length - diameter
            y_rab = y + self.box_height - mandrel_radius_ext - self.right_anchor
            center_point_rac = (x + self.reinforcement_length - mandrel_radius_ext,
                                y + self.box_height - mandrel_radius_ext)

            length_first_rect_bar -= mandrel_radius_ext if not self.bend_longitud else 0
            sides_first_rect_bar[1] = 0
            length_third_rect_bar -= mandrel_radius_ext
            sides_third_rect_bar = [1, 0, 1, 0]

            # From left to right.
            # Sixth bend curve.
            if not unifilar:
                steel_elements += curve(doc=document, center_point=center_point_rac,
                                        radius=self.mandrel_radius,
                                        start_angle=0,
                                        end_angle=90, thickness=diameter)

            # Second anchor rect bar (right).
            steel_elements += rect(doc=document,
                                   width=diameter,
                                   height=self.right_anchor, x=x_rab, y=y_rab,
                                   sides=sides_right_anchor)

        # Bending bar.
        if self.bend_longitud:

            alpha = rads(self.bend_angle / 2)
            dx1 = cos(alpha) * sin(alpha) * diameter * 2
            dy1 = sin(alpha) * sin(alpha) * diameter * 2
            dx2 = cos(alpha) * sin(alpha) * diameter * 4
            dy2 = sin(alpha) * sin(alpha) * diameter * 4

            longitud_mid = (self.reinforcement_length + self.bend_longitud) / 2
            longitud_curves = dx1 + dx2
            longitud_proyeccion = (self.bend_height + diameter - dy1 - dy2) / tan(alpha * 2)

            length_first_rect_bar -= (longitud_mid + longitud_curves + longitud_proyeccion)
            sides_first_rect_bar[1] = 0
            length_third_rect_bar -= (longitud_mid + longitud_curves + longitud_proyeccion)

            center_point_bc_first = (x_first_rect_bar + length_first_rect_bar,
                                     y + self.box_height - diameter * 2)
            center_point_bc_second = (x + (self.reinforcement_length - self.bend_longitud) / 2,
                                      y + self.box_height - self.bend_height)

            # From left to right.
            # First piece of rect bar.
            sides_first_rect_bar = sides_first_rect_bar if not unifilar else [1, 0, 0, 0]
            steel_elements += rect(doc=document,
                                   width=length_first_rect_bar,
                                   height=diameter,
                                   x=x_first_rect_bar,
                                   y=y + self.box_height - diameter,
                                   sides=sides_first_rect_bar)

            if not unifilar:
                # First curve of bend.
                steel_elements += curve(doc=document,
                                        center_point=center_point_bc_first,
                                        radius=diameter,
                                        start_angle=90 - self.bend_angle,
                                        end_angle=90,
                                        thickness=diameter)

            # First bend rect bar.
            if not unifilar:
                steel_elements += line(doc=document,
                                       p1=(x + (self.reinforcement_length - self.bend_longitud) / 2 - dx1,
                                           y + self.box_height - self.bend_height - diameter + dy1),
                                       p2=(x + (
                                               self.reinforcement_length - self.bend_longitud) / 2 - dx1 - longitud_proyeccion,
                                           y + self.box_height - dy2))

            steel_elements += line(doc=document,
                                   p1=(x + (self.reinforcement_length - self.bend_longitud) / 2 - dx2,
                                       y + self.box_height - self.bend_height - diameter * 2 + dy2),
                                   p2=(
                                       x + (
                                               self.reinforcement_length - self.bend_longitud) / 2 - dx2 - longitud_proyeccion,
                                       y + self.box_height - diameter - dy1))

            # Second curve of bend.
            if not unifilar:
                steel_elements += curve(doc=document,
                                        center_point=center_point_bc_second,
                                        radius=diameter,
                                        start_angle=270 - self.bend_angle,
                                        end_angle=270,
                                        thickness=diameter)

            # Second rect bar.
            sides_second_rect_bar = [1, 0, 1, 0] if not unifilar else [1, 0, 0, 0]
            steel_elements += rect(doc=document,
                                   width=self.bend_longitud,
                                   height=diameter,
                                   x=x + (self.reinforcement_length - self.bend_longitud) / 2,
                                   y=y + self.box_height - diameter * 2 - self.bend_height,
                                   sides=sides_second_rect_bar)

            # Third bend curve.
            if not unifilar:
                steel_elements += curve(doc=document,
                                        center_point=(x + longitud_mid,
                                                      y + self.box_height - self.bend_height),
                                        radius=diameter,
                                        start_angle=270,
                                        end_angle=270 + self.bend_angle,
                                        thickness=diameter)

            # Second bend bar.
            if not unifilar:
                steel_elements += line(doc=document,
                                       p1=(x + longitud_mid + dx1,
                                           y + self.box_height - self.bend_height - diameter + dy1),
                                       p2=(x + longitud_mid + dx1 + longitud_proyeccion,
                                           y + self.box_height - dy2))

            steel_elements += line(doc=document,
                                   p1=(x + longitud_mid + dx2,
                                       y + self.box_height - self.bend_height - diameter * 2 + dy2),
                                   p2=(x + longitud_mid + dx2 + longitud_proyeccion,
                                       y + self.box_height - diameter - dy1))

            # Fourth bend curve.
            if not unifilar:
                steel_elements += curve(doc=document,
                                        center_point=(x + longitud_mid + longitud_curves + longitud_proyeccion,
                                                      y + self.box_height - diameter * 2),
                                        radius=diameter,
                                        start_angle=90,
                                        end_angle=90 + self.bend_angle,
                                        thickness=diameter)

            # Third rect bar.
            sides_third_rect_bar = sides_third_rect_bar if not unifilar else [1, 0, 0, 0]
            steel_elements += rect(doc=document,
                                   width=length_third_rect_bar,
                                   height=diameter,
                                   x=x + longitud_mid + longitud_curves + longitud_proyeccion,
                                   y=y + self.box_height - diameter,
                                   sides=sides_third_rect_bar)

        else:
            # From left to right.
            # First rect bar (body).
            sides_first_rect_bar = sides_first_rect_bar if not unifilar else [1, 0, 0, 0]
            mandrel_radius_ext = 0 if not self.left_anchor and not self.right_anchor else mandrel_radius_ext
            steel_elements += rect(doc=document,
                                   width=length_first_rect_bar,
                                   height=diameter,
                                   x=x + mandrel_radius_ext,
                                   y=y + self.box_height - diameter,
                                   sides=sides_first_rect_bar)

        if dimensions:
            if self.left_anchor:
                dimension_elements += text(document=document,
                                           text="({:.2f})".format(self.left_anchor),
                                           height=settings["text_dim_height"],
                                           point=(x_lab - settings["text_dim_distance_vertical"],
                                                  y + self.left_anchor / 2),
                                           rotation=90,
                                           attr={"halign": 4, "valign": 0})

            if self.right_anchor:
                dimension_elements += text(document=document,
                                           text="({:.2f})".format(self.right_anchor),
                                           height=settings["text_dim_height"],
                                           point=(x_rab + settings["text_dim_distance_vertical"],
                                                  y + self.right_anchor / 2),
                                           rotation=90,
                                           attr={"halign": 4, "valign": 0})

            dimension_elements += text(document=document,
                                       text="({:.2f})".format(self.box_width),
                                       height=settings["text_dim_height"],
                                       point=(x + self.box_width / 2,
                                              y + self.box_height + settings["text_dim_distance_horizontal"]),
                                       rotation=0,
                                       attr={"halign": 4, "valign": 0})

        if denomination:
            denomination_elements += text(document=document,
                                          text=self.denomination,
                                          height=settings["text_denomination_height"],
                                          point=(x + self.box_width / 2,
                                                 y + self.box_height - settings["text_denomination_height"] - settings[
                                                     "text_denomination_distance"]),
                                          rotation=0,
                                          attr={"halign": 4, "valign": 0})

        # Setting groups of elements in dictionary.
        elements["steel_elements"] = steel_elements
        elements["dimension_elements"] = dimension_elements
        elements["denomination_elements"] = denomination_elements
        elements["text_elements"] = (elements["dimension_elements"] +
                                     elements["denomination_elements"])
        elements["all_elements"] = (elements["steel_elements"] +
                                    elements["dimension_elements"] +
                                    elements["denomination_elements"])

        if unifilar:
            translate(objects=elements["all_elements"],
                      vector=(0, -self.mandrel_radius_ext))

        # Orienting the bar (direction and orientation).
        self.__direc_orient(elements["all_elements"],
                            x=x,
                            y=y,
                            unifilar=unifilar)

        if self.orientation == Orientation.TOP:
            self.__direct_orient_text(elements["text_elements"])

        return elements

    # Drawing of transverse section of bar function.
    def draw_transverse(self,
                        document: Drawing,
                        x: float = None,
                        y: float = None,
                        settings: dict = BAR_SET_TRANSVERSE) -> dict:
        """
        Draws the transverse section of the bar in a DXF document.

        :param document: The DXF document to draw on.
        :type document: Drawing
        :param x: X coordinate for the drawing, defaults to self.x.
        :type x: float, optional
        :param y: Y coordinate for the drawing, defaults to self.y.
        :type y: float, optional
        :param settings: Dictionary of settings for dimensioning. Defaults to `BAR_SET_TRANSVERSE`.
        :type settings: dict, optional
        :return: Dict of drawing entities for the transverse section.
        :rtype: dict
        """
        if x is None:
            x = self.x

        if y is None:
            y = self.y

        elements = {}
        steel_elements = []

        # Checking if there is a defined transverse center.
        if self.transverse_center:
            x += self.transverse_center[0]
            y += self.transverse_center[1]

        # Drawing of circle.
        steel_elements += circle(doc=document,
                                 center_point=(x + self.radius, y + self.radius),
                                 radius=self.radius)

        # Setting groups of elements in dictionary.
        elements["steel_elements"] = steel_elements
        elements["all_elements"] = elements["steel_elements"]

        return elements

    # Function that orients drawing.
    def __direc_orient(self, group: list, x: float = None, y: float = None, unifilar: bool = False):
        """
        Adjusts the orientation of the drawing based on the direction and orientation of the bar.

        :param group: List of drawing entities to be oriented.
        :type group: list
        :param x: X coordinate for the orientation, defaults to self.x.
        :type x: float, optional
        :param y: Y coordinate for the orientation, defaults to self.y.
        :type y: float, optional
        :param unifilar: Whether to apply unifilar adjustments, defaults to False.
        :type unifilar: bool, optional
        """
        if x is None:
            x = self.x

        if y is None:
            y = self.y

        # Orienting the bar (direction and orientation).
        # Direction.
        if self.direction == Direction.VERTICAL:
            pivot_point = (x, y + self.box_height)
            vector_translate = (x - (pivot_point[0] * cos(pi / 2) - pivot_point[1] * sin(pi / 2)),
                                y - (pivot_point[0] * sin(pi / 2) + pivot_point[1] * cos(pi / 2)))

            rotate(group, pi / 2)
            translate(group, vector=vector_translate)

            if unifilar:
                translate(group, vector=(-self.diameter, 0))

        # Orientation.
        if self.orientation == Orientation.TOP:
            mirror(objects=group, mirror_type="x")
            if self.direction == Direction.VERTICAL:
                translate(group, vector=(0, self.box_width + y * 2))
            else:
                translate(group, vector=(0, self.box_height + y * 2))
                if unifilar:
                    translate(group, vector=(0, self.radius))

        elif self.orientation == Orientation.RIGHT:
            if self.direction == Direction.VERTICAL:
                if unifilar:
                    translate(group, vector=(-self.mandrel_radius_ext, 0))
            else:
                pass

        elif self.orientation == Orientation.LEFT:
            mirror(objects=group, mirror_type="y")

            if self.direction == Direction.VERTICAL:
                translate(group, vector=(self.box_height + x * 2, 0))
            else:
                translate(group, vector=(self.box_width + x * 2, 0))

    def __direct_orient_text(self, group: list, x: float = None, y: float = None):
        """
        Adjusts the text orientation in the drawing.

        :param group: List of drawing entities to be adjusted.
        :type group: list
        :param x: X coordinate for the text adjustment, defaults to self.x.
        :type x: float, optional
        :param y: Y coordinate for the text adjustment, defaults to self.y.
        :type y: float, optional
        """
        coordinates = [entitie.get_placement()[1] for entitie in group]
        mirror(group, mirror_type="x")

        for i, entitie in enumerate(group):
            entitie.set_placement([-coordinates[i][0], coordinates[i][1], coordinates[i][2]])

    def data(self) -> dict:
        """
        Collects and returns the essential attributes of the bar element in a dictionary format.

        :return: Dictionary containing key attributes of the bar such as denomination, length, diameter and weight.
        :rtype: dict
        """
        data = {"denomination": self.denomination,
                "length": self.length,
                "diameter": self.diameter,
                "weight": self.weight}

        return data

    def extract_data(self, labels: list[str] = None):
        """
        Extracts specific data attributes based on the provided list of labels.
        If no labels are provided, it defaults to extracting "denomination", "length", "diameter", and "weight".

        :param labels: List of attribute names to extract. Defaults to common attributes if not provided.
        :type labels: list[str], optional
        :return: A list of values corresponding to the requested labels.
        :rtype: list
        """
        if labels is None:
            labels = ["denomination", "length", "diameter", "weight"]

        data = self.data()

        data_required = []
        for key in labels:
            if key in data:
                data_required.append(data[key])
            else:
                data_required.append("-")

        return data_required



# Local imports.
from etacad.globals import DRotation
from etacad.geometry.polygon import Polygon
from etacad.globals import DRotation

# External imports.
import pytest


@pytest.fixture
def polygon_clockwise():
    return Polygon(vertices=[(0, 0), (0, 1), (1, 1), (2, 2), (3, 2), (3, 0)])


def test_polygon_clockwise_attributes(polygon_clockwise):
    assert polygon_clockwise.perimeter == 6.414213562373095
    assert polygon_clockwise.area == 4.5
    assert polygon_clockwise.centroid == (1.5, 1.0)


def test_polygon_clockwise_get_points_direction_of_rotation(polygon_clockwise):
    assert polygon_clockwise.get_points_direction_of_rotation() == DRotation.CLOCKWISE


def test_polygon_clockwise_get_points_between_index_limits_by_dr(polygon_clockwise):
    cw_list_1_4 = polygon_clockwise.get_points_between_index_limits_by_dr(index1=1,
                                                                          index2=4,
                                                                          direction_rotation=DRotation.CLOCKWISE)

    assert cw_list_1_4 == [(0, 1), (1, 1), (2, 2), (3, 2)]

    cw_list_0_5 = polygon_clockwise.get_points_between_index_limits_by_dr(index1=0,
                                                                          index2=5,
                                                                          direction_rotation=DRotation.CLOCKWISE)

    assert cw_list_0_5 == [(0, 0), (0, 1), (1, 1), (2, 2), (3, 2), (3, 0)]

    ccw_list_1_4 = polygon_clockwise.get_points_between_index_limits_by_dr(index1=1,
                                                                           index2=4,
                                                                           direction_rotation=DRotation.COUNTERCLOCKWISE)

    assert ccw_list_1_4 == [(0, 1), (0, 0), (3, 0), (3, 2)]

    ccw_list_0_4 = polygon_clockwise.get_points_between_index_limits_by_dr(index1=0,
                                                                           index2=4,
                                                                           direction_rotation=DRotation.COUNTERCLOCKWISE)

    assert ccw_list_0_4 == [(0, 0), (3, 0), (3, 2)]


def test_polygon_clockwise_get_side_normals(polygon_clockwise):
    sides = polygon_clockwise.get_side_equations()

    assert sides[0] == (1e+16, -0.0)
    assert sides[1] == (0, 1)
    assert sides[2] == (1, 0)
    assert sides[3] == (0, 2)
    assert sides[4] == (1e+16, -3e+16)


def test_polygon_clockwise_get_side_centers(polygon_clockwise):
    centers = polygon_clockwise.get_side_centers()

    assert centers[0] == (0, 0.5)
    assert centers[1] == (0.5, 1)
    assert centers[2] == (1.5, 1.5)
    assert centers[3] == (2.5, 2)
    assert centers[4] == (3, 1)


@pytest.fixture
def polygon_counterclockwise():
    return Polygon(vertices=[(3, 0), (3, 2), (2, 2), (1, 1), (0, 1), (0, 0)])


def test_polygon_counterclockwise_attributes(polygon_counterclockwise):
    assert polygon_counterclockwise.perimeter == 6.414213562373095
    assert polygon_counterclockwise.area == 4.5
    assert polygon_counterclockwise.centroid == (1.5, 1.0)


def test_polygon_counterclockwise_get_points_direction_of_rotation(polygon_counterclockwise):
    assert polygon_counterclockwise.get_points_direction_of_rotation() == DRotation.COUNTERCLOCKWISE


def test_polygon_counterclockwise_get_points_between_index_limits_by_dr(polygon_counterclockwise):
    cw_list_1_4 = polygon_counterclockwise.get_points_between_index_limits_by_dr(index1=1,
                                                                                 index2=4,
                                                                                 direction_rotation=DRotation.CLOCKWISE)

    assert cw_list_1_4 == [(3, 2), (3, 0), (0, 0), (0, 1)]

    cw_list_0_5 = polygon_counterclockwise.get_points_between_index_limits_by_dr(index1=0,
                                                                                 index2=5,
                                                                                 direction_rotation=DRotation.CLOCKWISE)

    assert cw_list_0_5 == [(3, 0), (0, 0)]

    ccw_list_1_4 = polygon_counterclockwise.get_points_between_index_limits_by_dr(index1=1,
                                                                                  index2=4,
                                                                                  direction_rotation=DRotation.COUNTERCLOCKWISE)

    assert ccw_list_1_4 == [(3, 2), (2, 2), (1, 1), (0, 1)]

    ccw_list_0_4 = polygon_counterclockwise.get_points_between_index_limits_by_dr(index1=0,
                                                                                  index2=4,
                                                                                  direction_rotation=DRotation.COUNTERCLOCKWISE)

    assert ccw_list_0_4 == [(3, 0), (3, 2), (2, 2), (1, 1), (0, 1)]





# -*- coding: utf-8 -*-

# Imports.
# Local imports.
from etacad.geometry.utils import get_lines_intersec
from etacad.drawing_utils import curve, dim_linear, line, mirror, rect_border_curve, rotate, text, translate
from etacad.globals import COS45, Direction, ElementTypes, Orientation, SIN45, STEEL_WEIGHT, STIRRUP_SET_TRANSVERSE

# External imports.
from attrs import define, field
from ezdxf.document import Drawing
from math import cos, sin, pi, floor


@define
class Stirrup:
    """
    Stirrup element, computes geometrics and physics props and manages dxf drawing methods.

    :param width: External width of stirrup.
    :type width: float
    :param height: External height of stirrup.
    :type height: float
    :param diameter: Diameter of stirrup bar.
    :type diameter: float
    :param reinforcement_length: Length of the stirrup reinforcement.
    :type reinforcement_length: float
    :param spacing: Stirrup spacing.
    :type spacing: float
    :param x: X coordinate of the bottom left corner Stirrup bounding box point.
    :type x: float
    :param y: Y coordinate of the bottom left corner Stirrup bounding box point.
    :type y: float
    :param mandrel_radius_top: Mandrel radius of stirrup bar at the top.
    :type mandrel_radius_top: float
    :param mandrel_radius_bottom: Mandrel radius of stirrup bar at the bottom.
    :type mandrel_radius_bottom: float
    :param anchor: Anchor length of stirrup.
    :type anchor: float
    :param direction: Direction of the stirrup (Horizontal or Vertical).
    :type direction: str
    :param orientation: Orientation of the stirrup (top, right, down, left).
    :type orientation: str

    :ivar reinforcement_length: Length of the stirrup reinforcement.
    :ivar spacing: Stirrup spacing.
    :ivar mandrel_radius_top: Mandrel radius of stirrup bar at the top.
    :ivar mandrel_radius_ext_top: Mandrel radius plus diameter of stirrup bar at the top.
    :ivar mandrel_radius_bottom: Mandrel radius of stirrup bar at the bottom.
    :ivar mandrel_radius_ext_bottom: Mandrel radius plus diameter of stirrup bar at the bottom.
    :ivar anchor: Anchor length of stirrup.
    :ivar quantity: Number of stirrups that forms de reinforcement.
    :ivar diameter: Diameter of stirrup bar.
    :ivar direction: Direction of the stirrup (Horizontal or Vertical).
    :ivar orientation: Orientation of the Stirrup (top, right, down, left).
    :ivar length: Length of the bar to fork the stirrup.
    :ivar weight: Weight of the stirrup, considering 7850 kg / m3.
    :ivar box_width: Width of the box that contains the stirrup.
    :ivar box_height: Height of the box that contains the stirrup.
    """
    # Stirrups attributes.
    width: float
    height: float
    diameter: float
    reinforcement_length: float
    spacing: float
    x: float = field(default=0)
    y: float = field(default=0)
    mandrel_radius_top: float = field(default=0)
    mandrel_radius_ext_top: float = field(init=False)
    mandrel_radius_bottom: float = field(default=0)
    mandrel_radius_ext_bottom: float = field(init=False)
    anchor: float = field(default=0)
    quantity: int = field(init=False)

    # Geometric attributes.
    length: float = field(init=False)
    direction: Direction = field(default=Direction.HORIZONTAL)
    orientation: Orientation = field(default=Orientation.BOTTOM)

    # Physics attributes.
    weight: float = field(init=False)

    # Box attributes.
    box_width: float = field(init=False)
    box_height: float = field(init=False)

    # Others.
    element_type: ElementTypes = field(default=ElementTypes.STIRRUP)
    denomination: str = field(default=None)
    position: str = field(default=None)

    def __attrs_post_init__(self):
        # Stirrups attributes.
        self.mandrel_radius_ext_top = self.mandrel_radius_top + self.diameter
        self.mandrel_radius_ext_bottom = self.mandrel_radius_bottom + self.diameter
        self.quantity = floor((self.reinforcement_length * 100) / (self.spacing * 100)) + 1

        # Geometric attributes.
        self.length = (self.width + self.height + self.anchor) * 2

        # Physics attributes.
        self.weight = (self.diameter ** 2 * pi / 4) * self.length * STEEL_WEIGHT

        # Box attributes.
        self.box_width = self.width
        self.box_height = self.height

    def draw_longitudinal(self,
                          document: Drawing,
                          x: float = None,
                          y: float = None,
                          unifilar=True) -> dict:
        """
        Draw the longitudinal reinforcement of the stirrup in the dxf file.

        :param document: Document in which it will be drawn.
        :type document: Drawing
        :param x: X coordinate of the bottom corner of the drawing.
        :type x: float
        :param y: Y coordinate of the bottom corner of the drawing.
        :type y: float
        :param unifilar: Single-line drawing.
        :type unifilar: bool

        :return: None.
        :rtype: None
        """
        if x is None:
            x = self.x

        if y is None:
            y = self.y

        elements = {}

        # Drawing stirrup steel bars.
        if unifilar:
            steel = [line(doc=document, p1=(x + self.spacing * i, y), p2=(x + self.spacing * i, y + self.height))[0]
                     for i in range(self.quantity)]
        else:
            steel = [line(doc=document, p1=(x + self.spacing * i, y), p2=(x + self.spacing * i, y + self.height))[0]
                     for i in range(self.quantity)]

        # Setting groups of elements in dictionary.
        elements["steel_elements"] = steel
        elements["all_elements"] = elements["steel_elements"]

        # Orienting the bar (direction and orientation).
        self.__direc_orient(elements["steel_elements"], x=x, y=y, longitudinal=True)

        return elements

    # Drawing transverse section of stirrup function.
    def draw_transverse(self,
                        document: Drawing,
                        x: float = None,
                        y: float = None,
                        unifilar: bool = False,
                        dimensions: bool = False,
                        settings: dict = STIRRUP_SET_TRANSVERSE) -> dict:
        """
        Draw the cross-section of the stirrup in the dxf file.

        :param document: Document in which it will be drawn.
        :type document: Drawing
        :param x: X coordinate of the bottom corner of the drawing.
        :type x: float
        :param y: Y coordinate of the bottom corner of the drawing.
        :type y: float
        :param unifilar: Single-line drawing.
        :type unifilar: bool
        :param dimensions: Dimensions drawing.
        :type dimensions: bool
        :param settings: Dictionary of settings for dimensioning. Defaults to `STIRRUP_SET_TRANSVERSE`.
        :type settings: dict, optional

        :return: None.
        :rtype: None
        """
        if x is None:
            x = self.x

        if y is None:
            y = self.y

        elements = {}
        steel_elements = []
        dim_elements = []

        # Curves radius of rect border bordes.
        curves_radius = [self.mandrel_radius_top, self.mandrel_radius_top, self.mandrel_radius_bottom,
                         self.mandrel_radius_bottom]

        # Constants equation of anchor lines.
        diff = 0 if unifilar else self.diameter
        m_45 = -1
        b_top_anchor_int = ((y + self.height - self.mandrel_radius_top - diff + self.mandrel_radius_top * SIN45)
                            - (m_45 * (x + self.mandrel_radius_top + diff + self.mandrel_radius_top * COS45)))
        b_top_anchor_ext = (
                (y + self.height - self.mandrel_radius_top - diff + (self.mandrel_radius_top + diff) * SIN45)
                - (m_45 * (x + self.mandrel_radius_top + diff + (self.mandrel_radius_top + diff) * COS45)))
        b_bottom_anchor_int = ((y + self.height - self.mandrel_radius_top - diff - self.mandrel_radius_top * SIN45)
                               - (m_45 * (x + self.mandrel_radius_top + diff - self.mandrel_radius_top * COS45)))
        b_bottom_anchor_ext = (
                (y + self.height - self.mandrel_radius_top - diff - (self.mandrel_radius_top + diff) * SIN45)
                - (m_45 * (x + self.mandrel_radius_top + diff - (self.mandrel_radius_top + diff) * COS45)))

        # Constants equation calculations of sides (top/left) lines.
        m_top_side = 0
        b_top_side = y + self.height if unifilar else y + self.height - self.diameter

        # Constants equation of 45° circle line.
        m_45_circle_line = 1
        if not unifilar:
            b_45_circle_line = ((y + self.box_height - self.mandrel_radius_ext_top)
                                - m_45_circle_line * (x + self.mandrel_radius_ext_top))
        else:
            b_45_circle_line = ((y + self.box_height - self.mandrel_radius_top)
                                - m_45_circle_line * (x + self.mandrel_radius_top))

        # Calculations.
        intersec_top_anchor_int = get_lines_intersec(m_45, b_top_anchor_int, m_45_circle_line, b_45_circle_line)[0]
        intersec_top_anchor_ext = get_lines_intersec(m_45, b_top_anchor_ext, m_top_side, b_top_side)[0]
        intersect_bottom_anchor_int = get_lines_intersec(m_45, b_bottom_anchor_int, m_45_circle_line, b_45_circle_line)[
            0]
        intersect_bottom_anchor_ext = get_lines_intersec(m_45, b_bottom_anchor_ext, m_45_circle_line, b_45_circle_line)[
            0]

        if unifilar:
            # Calculations.
            p1_top_anchor_int = (intersec_top_anchor_int,
                                 intersec_top_anchor_int * m_45 + b_top_anchor_int)
            p2_top_anchor_int = (p1_top_anchor_int[0] + COS45 * self.anchor,
                                 p1_top_anchor_int[1] - SIN45 * self.anchor)

            p1_bottom_anchor_int = (intersect_bottom_anchor_int,
                                    intersect_bottom_anchor_int * m_45 + b_bottom_anchor_int)
            p2_bottom_anchor_int = (p1_bottom_anchor_int[0] + COS45 * self.anchor,
                                    p1_bottom_anchor_int[1] - SIN45 * self.anchor)

            p1_bottom_anchor_ext = p1_bottom_anchor_int
            p2_bottom_anchor_ext = p2_bottom_anchor_int

            # Anchor drawing.
            steel_elements += line(doc=document, p1=p1_top_anchor_int,
                                   p2=p2_top_anchor_int)  # Internal line top anchor.
            steel_elements += line(doc=document, p1=p1_bottom_anchor_int,
                                   p2=p2_bottom_anchor_int)  # Internal line bot anchor.

            # Rectangle drawing (circle borders).
            steel_elements += rect_border_curve(doc=document, width=self.width, height=self.height,
                                                radius=self.mandrel_radius_top, x=x, y=y, thickness=0,
                                                sides=[1, 1, 1, 1],
                                                curves_radius=curves_radius)

        else:
            # Drawing of side borders.
            steel_elements = rect_border_curve(doc=document, width=self.width - 2 * self.diameter,
                                               height=self.height - 2 * self.diameter, radius=self.mandrel_radius_top,
                                               x=x + self.diameter, y=y + self.diameter, thickness=self.diameter,
                                               sides=[1, 1, 1, 0], curves=[0, 1, 1, 1], curves_radius=curves_radius)

            # External left side border.
            steel_elements += line(doc=document, p1=(x, y + self.mandrel_radius_ext_bottom),
                                   p2=(x, y + self.box_height - self.mandrel_radius_ext_top))

            # Drawing of anchors.
            # Calculatios of points.
            p1_top_anchor_int = (intersec_top_anchor_int,
                                 intersec_top_anchor_int * m_45 + b_top_anchor_int)
            p2_top_anchor_int = (p1_top_anchor_int[0] + COS45 * self.anchor,
                                 p1_top_anchor_int[1] - SIN45 * self.anchor)
            p1_top_anchor_ext = (intersec_top_anchor_ext,
                                 intersec_top_anchor_ext * m_45 + b_top_anchor_ext)
            p2_top_anchor_ext = (p2_top_anchor_int[0] + self.diameter * COS45,
                                 p2_top_anchor_int[1] + self.diameter * SIN45)
            p1_bottom_anchor_int = (intersect_bottom_anchor_int,
                                    m_45 * intersect_bottom_anchor_int + b_bottom_anchor_int)
            p2_bottom_anchor_int = (p1_bottom_anchor_int[0] + COS45 * self.anchor,
                                    p1_bottom_anchor_int[1] - SIN45 * self.anchor)
            p1_bottom_anchor_ext = (intersect_bottom_anchor_ext,
                                    m_45 * intersect_bottom_anchor_ext + b_bottom_anchor_ext)
            p2_bottom_anchor_ext = (p1_bottom_anchor_ext[0] + COS45 * self.anchor,
                                    p1_bottom_anchor_ext[1] - SIN45 * self.anchor)

            # Center of top left curve.
            center_point_curve = (x + self.mandrel_radius_ext_top, y + self.box_height - self.mandrel_radius_ext_top)

            # Drawing of lines anchors.
            steel_elements += line(doc=document, p1=p1_top_anchor_ext, p2=p2_top_anchor_ext)
            steel_elements += line(doc=document, p1=p1_top_anchor_int, p2=p2_top_anchor_int)
            steel_elements += line(doc=document, p1=p1_bottom_anchor_int, p2=p2_bottom_anchor_int)
            steel_elements += line(doc=document, p1=p1_bottom_anchor_ext, p2=p2_bottom_anchor_ext)

            steel_elements += line(doc=document, p1=p2_bottom_anchor_int,
                                   p2=p2_bottom_anchor_ext)  # Closed lines bottom.
            steel_elements += line(doc=document, p1=p2_top_anchor_int, p2=p2_top_anchor_ext)  # Closed lines top.

            # Left internal side.
            steel_elements += line(doc=document, p1=(x + self.diameter, y + self.mandrel_radius_ext_bottom),
                                   p2=(p1_top_anchor_ext[0] - (self.diameter * 2 + self.mandrel_radius_top * 2) * COS45,
                                       p1_top_anchor_ext[1] - (
                                               self.diameter * 2 + self.mandrel_radius_top * 2) * SIN45))

            # Top left curve.
            steel_elements += curve(doc=document, center_point=center_point_curve, start_angle=90, end_angle=225,
                                    radius=self.mandrel_radius_top + self.diameter, thickness=0)  # Internal.
            steel_elements += curve(doc=document, center_point=center_point_curve, start_angle=45, end_angle=225,
                                    radius=self.mandrel_radius_top, thickness=0)  # External.

        if dimensions:
            # Anchor dimension.
            dim_elements += dim_linear(document=document,
                                       p_base=(x + self.width / 2 - settings["text_dim_distance_anchor"],
                                               y + self.height - settings["text_dim_distance_anchor"]),
                                       p1=p1_bottom_anchor_ext,
                                       p2=p2_bottom_anchor_ext, rotation=315,
                                       dimstyle=settings["dim_style"])

            # Vertical dimension.
            dim_elements += dim_linear(document=document,
                                       p_base=(x - settings["text_dim_distance_vertical"], y + self.height / 2),
                                       p1=(x, y),
                                       p2=(x, y + self.height),
                                       rotation=90,
                                       dimstyle=settings["dim_style"])

            # Anchor horizontal.
            dim_elements += dim_linear(document=document,
                                       p_base=(x + self.width / 2,
                                               y + self.height + settings["text_dim_distance_horizontal"]),
                                       p1=(x, y + self.height),
                                       p2=(x + self.width, y + self.height),
                                       dimstyle=settings["dim_style"])

            # Text for length count.
            dim_elements += text(document=document,
                                 text="L: {0:.2f}".format(self.length),
                                 height=settings["text_length_count_height"],
                                 point=(x + self.width / 2, y - settings["text_distance_length_count"]),
                                 attr={"halign": 4, "valign": 0})

        # Setting groups of elements in dictionary.
        elements["steel_elements"] = steel_elements
        elements["dimensions_elements"] = dim_elements
        elements["all_elements"] = (elements["steel_elements"] +
                                    elements["dimensions_elements"])

        # Orienting the bar (direction and orientation).
        self.__direc_orient(elements["steel_elements"], x=x, y=y, longitudinal=False)

        return elements

    # Function that orients drawing.
    def __direc_orient(self,
                       group: list,
                       x: float = None,
                       y: float = None,
                       longitudinal: bool = False):
        """
        Orient the elements according to the parameters of the stirrup in the group entered.

        :param group: Group of elements dxf to orient.
        :param x: X coordinate of the bottom corner of the drawing.
        :type x: float
        :param y: Y coordinate of the bottom corner of the drawing.
        :type y: float
        :param longitudinal: Type of drawing to which the dxf elements belong (longitudinal or transverse).
        :type longitudinal: bool

        :return: Group of elements dxf to oriented.
        :rtype: list
        """
        # Orienting the bar (direction and orientation).
        # Direction.
        if self.direction == Direction.VERTICAL:
            pivot_point = (x, y + self.box_height)
            vector_translate = (x - (pivot_point[0] * cos(pi / 2) - pivot_point[1] * sin(pi / 2)),
                                y - (pivot_point[0] * sin(pi / 2) + pivot_point[1] * cos(pi / 2)))

            rotate(group, pi / 2)
            translate(group, vector=vector_translate)

        # Orientation.
        if self.orientation == Orientation.LEFT:
            mirror(objects=group, mirror_type="y")

            if self.direction == Direction.VERTICAL:
                translate(group, vector=(self.box_height + x * 2, 0))
            else:
                if longitudinal:
                    translate(group, vector=(self.reinforcement_length + x * 2, 0))
                else:
                    translate(group, vector=(self.box_width + x * 2, 0))

    def data(self) -> dict:
        """
        Collects and returns the essential attributes of the stirrup element in a dictionary format.

        :return: Dictionary containing key attributes of the bar such as denomination, length, diameter, weight and
        quantity.
        :rtype: dict
        """
        data = {"denomination": self.denomination,
                "length": self.length,
                "diameter": self.diameter,
                "weight": self.weight,
                "quantity": self.quantity}

        return data

    def extract_data(self, labels: list[str] = None) -> list:
        """
        Extracts specific data attributes based on the provided list of labels.
        If no labels are provided, it defaults to extracting "denomination", "length", "diameter", "weight" and
        quantity.

        :param labels: List of attribute names to extract. Defaults to common attributes if not provided.
        :type labels: list[str], optional

        :return: A list of values corresponding to the requested labels.
        :rtype: list
        """
        if labels is None:
            labels = ["denomination", "length", "diameter", "weight", "quantity"]

        data = self.data()

        data_required = []
        for key in labels:
            if key in data:
                data_required.append(data)
            else:
                data_required.append("-")

        return data_required
